from __future__ import annotations

from itertools import combinations

from plagdef.model.extension import ClusterBuilder
from plagdef.model.filtering import ClusterFilter
from plagdef.model.models import Document, Match, DocumentPairMatches, Cluster, Fragment
from plagdef.model.preprocessing import Preprocessor
from plagdef.model.seeding import SeedFinder


class DocumentMatcher:
    def __init__(self, config: dict):
        self._preprocessor = Preprocessor(config['min_sent_len'], config['rem_stop_words'])
        self._seeder = SeedFinder(config['min_cos_sim'], config['min_dice_sim'])
        self._extender = ClusterBuilder(config['adjacent_sents_gap'], config['min_adjacent_sents_gap'],
                                        config['min_sent_number'], config['min_cluster_cos_sim'])
        self._cluster_filter = ClusterFilter(config['min_cluster_char_len'])
        self._adjacent_sents_gap_summary = config['adjacent_sents_gap_summary']
        self._min_verbatim_match_char_len = config['min_verbatim_match_char_len']

    def find_matches(self, lang: str, docs: list[Document], common_docs=None) -> set[DocumentPairMatches]:
        self._preprocessor.preprocess(lang, docs, common_docs)
        matches = set()
        for doc1, doc2 in combinations(docs, 2):
            seeds = self._seeder.seed(doc1, doc2)
            clusters = self._extender.extend(seeds)
            clusters = self._cluster_filter.filter(clusters)
            verbatim_matches = self._verbatim_matches(clusters)
            if len(verbatim_matches):  # Plagiarism type: verbatim
                matches.add(DocumentPairMatches(verbatim_matches))
            else:
                summary_clusters = self._extender.extend(seeds, self._adjacent_sents_gap_summary)
                summary_clusters = self._cluster_filter.filter(summary_clusters)
                if len(summary_clusters):
                    sum_cluster_len_doc1, sum_cluster_len_doc2 = \
                        tuple(map(sum, zip(*(cluster.char_lengths() for cluster in summary_clusters))))

                    if sum_cluster_len_doc1 >= 3 * sum_cluster_len_doc2 \
                        or sum_cluster_len_doc2 >= 3 * sum_cluster_len_doc1:  # Plagiarism type: summary
                        matches.add(DocumentPairMatches({Match.from_cluster(cluster) for cluster in summary_clusters}))
                    elif len(clusters):  # Plagiarism type: intelligent
                        matches.add(DocumentPairMatches({Match.from_cluster(cluster) for cluster in clusters}))
        return matches

    def _verbatim_matches(self, clusters: set[Cluster]) -> set[Match]:
        matches = set()
        for cluster in clusters:
            cluster_matches = self._common_words(cluster)
            matches.update(self._resolve_overlaps(cluster_matches))
        return matches

    def _common_words(self, cluster: Cluster) -> set[Match]:
        verbatim_matches = set()
        frag1_words = [word for sent_words in [sent.words for sent in cluster.sents_doc1] for word in sent_words]
        frag2_words = [word for sent_words in [sent.words for sent in cluster.sents_doc2] for word in sent_words]
        lookup = [[0 for _ in range(len(frag2_words) + 1)] for _ in range(len(frag1_words) + 1)]
        for i in range(1, len(frag1_words) + 1):
            for j in range(1, len(frag2_words) + 1):
                if frag1_words[i - 1].text.lower() == frag2_words[j - 1].text.lower():
                    lookup[i][j] = lookup[i - 1][j - 1] + 1
                    match_char_len = sum([len(word) for word in frag1_words[i - lookup[i][j]:i]])
                    if match_char_len >= self._min_verbatim_match_char_len:
                        frag1_first, frag2_first = i - lookup[i][j], j - lookup[i][j]
                        frag1 = Fragment(frag1_words[frag1_first].start_char, frag1_words[i - 1].end_char, cluster.doc1)
                        frag2 = Fragment(frag2_words[frag2_first].start_char, frag2_words[j - 1].end_char, cluster.doc2)
                        verbatim_matches.add(Match(frag1, frag2))
        return verbatim_matches

    def _resolve_overlaps(self, matches: set[Match]) -> set[Match]:
        non_ol_matches = set()
        for match in sorted(matches, key=len, reverse=True):
            if not any(match.overlaps_with(non_ol_match) for non_ol_match in non_ol_matches):
                non_ol_matches.add(match)
        return non_ol_matches



from __future__ import annotations

from collections import Counter

import stanza
from stanza import Pipeline

from plagdef.model import stopwords
from plagdef.model.models import Document, Sentence, Word

PRCS = 'tokenize,mwt,pos,lemma'
PIPE_LVL = 'WARN'
LOAD_LVL = 'INFO'
LANG_CODES = {'ger': 'de', 'eng': 'en'}


class Preprocessor:
    def __init__(self, min_sent_len: int, rem_stop_words: bool):
        self._min_sent_len = min_sent_len
        self._rem_stop_words = rem_stop_words

    def preprocess(self, lang: str, docs: list[Document], common_docs: list[Document] = None):
        if common_docs:
            self.preprocess(lang, common_docs)
        nlp_model = _nlp_pipe(lang)
        parsed_docs = nlp_model([stanza.Document([], text=doc.text) for doc in docs]) if docs else []
        common_sent_bows = [sent.bow for doc_sents in [doc.sents for doc in common_docs] for sent in doc_sents] \
            if common_docs else []
        stop_words = stopwords.ENGLISH if lang == 'eng' else stopwords.GERMAN
        for doc_idx, parsed_doc in enumerate(parsed_docs):
            for sent_idx, sent in enumerate(parsed_doc.sentences):
                non_punct_words = [word for word in sent.words if not word.upos == 'PUNCT']
                if self._rem_stop_words:
                    sent_lemmas = [word.lemma for word in non_punct_words if word.text.lower() not in stop_words]
                else:
                    sent_lemmas = [word.lemma for word in non_punct_words]
                if len(sent_lemmas):
                    lemma_count = Counter(sent_lemmas)
                    sentence = Sentence(sent.tokens[0].start_char, sent.tokens[-1].end_char, lemma_count, docs[doc_idx])
                    sentence.words = _to_words(sent.tokens, sentence)
                    if sentence.bow not in common_sent_bows:
                        docs[doc_idx].sents.add(sentence)
                        for lemma in lemma_count.keys():
                            docs[doc_idx].vocab[lemma] += 1
            self._join_small_sentences(docs[doc_idx])

    def _join_small_sentences(self, doc: Document):
        idx, sent_count = 0, len(doc.sents)
        while idx < sent_count - 1:
            sent1, sent2 = doc.sents[idx], doc.sents[idx + 1]
            if sum(sent1.bow.values()) < self._min_sent_len \
                or (sent2 == doc.sents[-1] and sum(sent2.bow.values()) < self._min_sent_len):
                for lemma in sent1.bow.keys():
                    if lemma in sent2.bow:
                        doc.vocab[lemma] -= 1
                joined_sent = Sentence(sent1.start_char, sent2.end_char, sent1.bow + sent2.bow, doc)
                joined_sent.words = sent1.words + sent2.words
                doc.sents.remove(sent1), doc.sents.remove(sent2)
                doc.sents.add(joined_sent)
                sent_count -= 1
            idx += 1


def _nlp_pipe(lang: str) -> Pipeline:
    if lang in LANG_CODES:
        try:
            return stanza.Pipeline(LANG_CODES[lang], processors=PRCS, logging_level=PIPE_LVL)
        except:  # Unpickling error raises Exception, cannot narrow
            stanza.download(LANG_CODES[lang], processors=PRCS, logging_level=LOAD_LVL)
            return stanza.Pipeline(LANG_CODES[lang], processors=PRCS, logging_level=PIPE_LVL)
    else:
        raise UnsupportedLanguageError(f'The language "{lang}" is currently not supported.')


def _to_words(stanza_tokens, sentence: Sentence) -> list[Word]:
    words = []
    for stanza_token in stanza_tokens:
        for stanza_word in stanza_token.words:
            if not stanza_word.upos == 'PUNCT':
                words.append(Word(stanza_token.start_char, stanza_token.end_char, sentence))
    return words


class UnsupportedLanguageError(Exception):
    pass
